@@ -18,11 +18,14 @@
 -- we represent the notion of a relative root by "@.@". The relative root denotes
 -- the directory which contains the first component of a relative path.
 
-{-# LANGUAGE TemplateHaskell #-}
-{-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE EmptyDataDecls #-}
 {-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE TemplateHaskell #-}
 
 module Path.PLATFORM_NAME
   (-- * Types
@@ -31,6 +34,7 @@ module Path.PLATFORM_NAME
   ,Rel
   ,File
   ,Dir
+  ,SomeBase(..)
    -- * Exceptions
   ,PathException(..)
   -- * QuasiQuoters
@@ -51,6 +55,7 @@ module Path.PLATFORM_NAME
   ,(</>)
   ,stripProperPrefix
   ,isProperPrefixOf
+  ,replaceProperPrefix
   ,parent
   ,filename
   ,dirname
@@ -58,17 +63,23 @@ module Path.PLATFORM_NAME
   ,splitExtension
   ,fileExtension
   ,replaceExtension
+  ,mapSomeBase
+  ,prjSomeBase
    -- * Parsing
   ,parseAbsDir
   ,parseRelDir
   ,parseAbsFile
   ,parseRelFile
+  ,parseSomeDir
+  ,parseSomeFile
   -- * Conversion
   ,toFilePath
   ,fromAbsDir
   ,fromRelDir
   ,fromAbsFile
   ,fromRelFile
+  ,fromSomeDir
+  ,fromSomeFile
   -- * TemplateHaskell constructors
   -- | These require the TemplateHaskell language extension.
   ,mkAbsDir
@@ -86,36 +97,40 @@ module Path.PLATFORM_NAME
   )
   where
 
+import           Control.Applicative (Alternative(..))
+import           Control.DeepSeq (NFData (..))
 import           Control.Exception (Exception(..))
 import           Control.Monad (liftM, when)
 import           Control.Monad.Catch (MonadThrow(..))
-import           Data.Aeson (FromJSON (..), FromJSONKey(..))
+import           Data.Aeson (FromJSON (..), FromJSONKey(..), ToJSON(..))
 import qualified Data.Aeson.Types as Aeson
 import           Data.Data
 import qualified Data.Text as T
-import           Data.List
+import           Data.Hashable
+import qualified Data.List as L
 import           Data.Maybe
+import           GHC.Generics (Generic)
 import           Language.Haskell.TH
 import           Language.Haskell.TH.Syntax (lift)
 import           Language.Haskell.TH.Quote (QuasiQuoter(..))
-import           Path.Internal
+import           Path.Internal.PLATFORM_NAME
 import qualified System.FilePath.PLATFORM_NAME as FilePath
 
 --------------------------------------------------------------------------------
 -- Types
 
 -- | An absolute path.
-data Abs deriving (Typeable)
+data Abs deriving (Typeable, Data)
 
 -- | A relative path; one without a root. Note that a @..@ path component to
 -- represent the parent directory is not allowed by this library.
-data Rel deriving (Typeable)
+data Rel deriving (Typeable, Data)
 
 -- | A file path.
-data File deriving (Typeable)
+data File deriving (Typeable, Data)
 
 -- | A directory path.
-data Dir deriving (Typeable)
+data Dir deriving (Typeable, Data)
 
 instance FromJSON (Path Abs File) where
   parseJSON = parseJSONWith parseAbsFile
@@ -176,6 +191,8 @@ data PathException
   | InvalidRelDir FilePath
   | InvalidAbsFile FilePath
   | InvalidRelFile FilePath
+  | InvalidFile FilePath
+  | InvalidDir FilePath
   | NotAProperPrefix FilePath FilePath
   | HasNoExtension FilePath
   | InvalidExtension String
@@ -307,7 +324,7 @@ infixr 5 </>
 stripProperPrefix :: MonadThrow m
          => Path b Dir -> Path b t -> m (Path Rel t)
 stripProperPrefix (Path p) (Path l) =
-  case stripPrefix p l of
+  case L.stripPrefix p l of
     Nothing -> throwM (NotAProperPrefix p l)
     Just "" -> throwM (NotAProperPrefix p l)
     Just ok -> return (Path ok)
@@ -325,6 +342,15 @@ stripProperPrefix (Path p) (Path l) =
 isProperPrefixOf :: Path b Dir -> Path b t -> Bool
 isProperPrefixOf p l = isJust (stripProperPrefix p l)
 
+-- | Change from one directory prefix to another.
+--
+-- Throw 'NotAProperPrefix' if the first argument is not a proper prefix of the
+-- path.
+--
+-- >>> replaceProperPrefix $(mkRelDir "foo") $(mkRelDir "bar") $(mkRelFile "foo/file.txt") == $(mkRelFile "bar/file.txt")
+replaceProperPrefix :: MonadThrow m => Path b Dir -> Path b' Dir -> Path b t -> m (Path b' t)
+replaceProperPrefix src dst fp = (dst </>) <$> stripProperPrefix src fp
+
 -- | Take the parent path component from a path.
 --
 -- The following properties hold:
@@ -795,6 +821,99 @@ normalizeFilePath
   | IS_WINDOWS = normalizeWindowsSeps . FilePath.normalise
   | otherwise  = normalizeLeadingSeps . FilePath.normalise
 
+-- | Path of some type.  @t@ represents the type, whether file or
+-- directory.  Pattern match to find whether the path is absolute or
+-- relative.
+data SomeBase t = Abs (Path Abs t)
+                | Rel (Path Rel t)
+  deriving (Typeable, Generic, Eq, Ord)
+
+instance NFData (SomeBase t) where
+  rnf (Abs p) = rnf p
+  rnf (Rel p) = rnf p
+
+instance Show (SomeBase t) where
+  show = show . fromSomeBase
+
+instance ToJSON (SomeBase t) where
+  toJSON = toJSON . fromSomeBase
+  {-# INLINE toJSON #-}
+#if MIN_VERSION_aeson(0,10,0)
+  toEncoding = toEncoding . fromSomeBase
+  {-# INLINE toEncoding #-}
+#endif
+
+instance Hashable (SomeBase t) where
+  -- See 'Hashable' 'Path' instance for details.
+  hashWithSalt n path = hashWithSalt n (fromSomeBase path)
+
+instance FromJSON (SomeBase Dir) where
+  parseJSON = parseJSONWith parseSomeDir
+  {-# INLINE parseJSON #-}
+
+instance FromJSON (SomeBase File) where
+  parseJSON = parseJSONWith parseSomeFile
+  {-# INLINE parseJSON #-}
+
+-- | Helper to project the contents out of a SomeBase object.
+--
+-- >>> prjSomeBase toFilePath (Abs [absfile|/foo/bar/cow.moo|]) == "/foo/bar/cow.moo"
+--
+prjSomeBase :: (forall b . Path b t -> a) -> SomeBase t -> a
+prjSomeBase f = \case
+  Abs a -> f a
+  Rel r -> f r
+
+-- | Helper to apply a function to the SomeBase object
+--
+-- >>> mapSomeBase parent (Abs [absfile|/foo/bar/cow.moo|]) == Abs [absdir|"/foo/bar"|]
+--
+mapSomeBase :: (forall b . Path b t -> Path b t') -> SomeBase t -> SomeBase t'
+mapSomeBase f = \case
+  Abs a -> Abs $ f a
+  Rel r -> Rel $ f r
+
+-- | Convert a valid path to a 'FilePath'.
+fromSomeBase :: SomeBase t -> FilePath
+fromSomeBase = prjSomeBase toFilePath
+
+-- | Convert a valid directory to a 'FilePath'.
+fromSomeDir :: SomeBase Dir -> FilePath
+fromSomeDir = fromSomeBase
+
+-- | Convert a valid file to a 'FilePath'.
+fromSomeFile :: SomeBase File -> FilePath
+fromSomeFile = fromSomeBase
+
+-- | Convert an absolute or relative 'FilePath' to a normalized 'SomeBase'
+-- representing a directory.
+--
+-- Throws: 'InvalidDir' when the supplied path:
+--
+-- * contains a @..@ path component representing the parent directory
+-- * is not a valid path (See 'FilePath.isValid')
+parseSomeDir :: MonadThrow m => FilePath -> m (SomeBase Dir)
+parseSomeDir fp = maybe (throwM (InvalidDir fp)) pure
+                $ (Abs <$> parseAbsDir fp)
+              <|> (Rel <$> parseRelDir fp)
+
+-- | Convert an absolute or relative 'FilePath' to a normalized 'SomeBase'
+-- representing a file.
+--
+-- Throws: 'InvalidFile' when the supplied path:
+--
+-- * is a directory path i.e.
+--
+--     * has a trailing path separator
+--     * is @.@ or ends in @/.@
+--
+-- * contains a @..@ path component representing the parent directory
+-- * is not a valid path (See 'FilePath.isValid')
+parseSomeFile :: MonadThrow m => FilePath -> m (SomeBase File)
+parseSomeFile fp = maybe (throwM (InvalidFile fp)) pure
+                 $ (Abs <$> parseAbsFile fp)
+               <|> (Rel <$> parseRelFile fp)
+
 --------------------------------------------------------------------------------
 -- Deprecated
 
